"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.registerEventListeners = exports.refreshSeenSequences = exports.checkForMissedVaas = void 0;
const application_1 = require("../../application");
const utils_1 = require("../../utils");
const helpers_1 = require("./helpers");
const storage_1 = require("./storage");
async function checkForMissedVaas(filter, redis, processVaa, opts, prefix, previousSafeSequence, logger) {
    const { emitterChain, emitterAddress } = filter;
    const startingSeqConfig = opts.startingSequenceConfig?.[emitterChain];
    const startingIndex = await (0, storage_1.calculateStartingIndex)(redis, prefix, emitterChain, emitterAddress, previousSafeSequence);
    const seenSequences = await (0, storage_1.getAllProcessedSeqsInOrder)(redis, prefix, emitterChain, emitterAddress, startingIndex, startingSeqConfig);
    const processed = [];
    const failedToRecover = [];
    const failedToReprocess = [];
    let missingSequences = [];
    if (seenSequences.length) {
        const first = seenSequences[0];
        const last = seenSequences[seenSequences.length - 1];
        logger?.info(`Scanning sequences from ${first} to ${last} for missing sequences`);
        // Check if there is any leap between the sequences seen,
        // and try reprocessing them if any:
        missingSequences = scanForSequenceLeaps(seenSequences);
        await (0, utils_1.mapConcurrent)(missingSequences, async (sequence) => {
            const vaa = {
                ...filter,
                sequence: sequence.toString(),
            };
            const seqString = sequence.toString();
            let vaaResponse;
            try {
                vaaResponse = await (0, helpers_1.tryFetchVaa)(vaa, opts.wormholeRpcs, opts.fetchVaaRetries);
                if (!vaaResponse) {
                    // this is a sequence that we found in the middle of two other sequences we processed,
                    // so we can consider this VAA not existing an error.
                    throw new Error("VAA Sequence not found.");
                }
            }
            catch (error) {
                // We have already retried a few times. We'll swallow the error and mark
                // the VAA as failed and seen.
                // VAAs marked as failed generate a metric that can be used to trigger an alert
                // for developers to take a closer look.
                // At the time of the implementation of this worker, the guardian network in testnet has
                // only one guardian, and this makes it so that it's possible for some VAAs to be missed
                // by the guardian.
                // Right now the VAAs marked as failed are logged on each iteration of the missed VAA check
                // but it would be nice to have a way to query them
                logger?.error(`Error fetching VAA for missing sequence ${sequence}`, error);
                failedToRecover.push(seqString);
                return;
            }
            try {
                await processVaa(Buffer.from(vaaResponse.vaaBytes));
                logger?.debug(`Recovered missing VAA ${seqString}`);
                processed.push(seqString);
            }
            catch (error) {
                // If we succeeded to fetch the VAA The error to reprocess is in our side (eg: redis failed)
                // We won't mark it as failed, so that it's retried on the next run of the missed VAA Worker.
                // We won't throw the error so that other vaas can be processed, but we'll add it to the list
                // of "failedToReprocess" so that we can log it and alert on it.
                // If you see this log while troubleshooting, it probably means that there is an issue on the
                // relayer side since the VAA was successfully fetched but failed to be processed.
                logger?.error(`Failed to reprocess vaa found missing. ${seqString}`, error);
                failedToReprocess.push(seqString);
            }
        }, opts.vaasFetchConcurrency);
        if (failedToRecover.length)
            await (0, storage_1.batchMarkAsFailedToRecover)(redis, prefix, emitterChain, emitterAddress, failedToRecover);
        const allSeenVaas = processed.concat(failedToRecover);
        if (allSeenVaas.length)
            await (0, storage_1.batchMarkAsSeen)(redis, prefix, emitterChain, emitterAddress, allSeenVaas);
    }
    // look ahead of greatest seen sequence in case the next vaa was missed
    // continue looking ahead until a vaa can't be fetched
    const lastSeq = seenSequences[seenSequences.length - 1]
        ? seenSequences[seenSequences.length - 1] + 1n
        : null;
    let lookAheadSequence = lastSeq && startingSeqConfig
        ? lastSeq > startingSeqConfig
            ? lastSeq
            : startingSeqConfig // same as Math.max, which doesn't support bigint
        : lastSeq || startingSeqConfig;
    logger?.info(`Looking ahead for missed VAAs from sequence: ${lookAheadSequence}`);
    const lookAheadSequences = [];
    if (lookAheadSequence) {
        for (let seq = lookAheadSequence; true; seq++) {
            const vaaKey = {
                ...filter,
                sequence: seq.toString(),
            };
            let vaa = null;
            try {
                vaa = await (0, helpers_1.tryFetchVaa)(vaaKey, opts.wormholeRpcs, 3);
            }
            catch (error) {
                logger?.error(`Error FETCHING Look Ahead VAA. Sequence ${seq}. Error: `, error);
            }
            if (!vaa)
                break;
            lookAheadSequences.push(seq.toString());
            logger?.info(`Found Look Ahead VAA. Sequence: ${seq.toString()}`);
            try {
                // since we add this VAA to the queue, there's no need to mark it as seen
                // (it will be automatically marked as seen when the "added" event is fired)
                await processVaa(Buffer.from(vaa.vaaBytes));
                lookAheadSequence = seq + 1n;
                processed.push(seq.toString());
            }
            catch (error) {
                logger?.error(`Error PROCESSING Look Ahead VAA. Sequence: ${seq.toString()}. Error:`, error);
            }
        }
    }
    else {
        logger?.warn(`No VAAs seen and no starting sequence was configured. Won't look ahead for missed VAAs.`);
    }
    return {
        processed,
        failedToRecover,
        failedToReprocess,
        lookAheadSequences,
        seenSequences: seenSequences.map((s) => s.toString()),
        missingSequences: missingSequences.map(s => s.toString()),
    };
}
exports.checkForMissedVaas = checkForMissedVaas;
async function refreshSeenSequences(redisPool, filters, opts) {
    await redisPool.use(async (redis) => {
        if (opts.storagePrefix && opts.forceSeenKeysReindex) {
            await (0, storage_1.deleteExistingSeenVAAsData)(filters, redis, opts.storagePrefix, opts.logger);
        }
        // The prefix used by the storage to store workflows.
        // We'll use it to go into the queue, and build a map of what
        // sequences have already been processed according to the store
        if (opts.storagePrefix) {
            const startTime = Date.now();
            const scannedKeys = await (0, storage_1.updateSeenSequences)(filters, redis, opts.storagePrefix);
            const elapsedTime = Date.now() - startTime;
            opts.logger?.info(`Scanned ${scannedKeys} keys in ${elapsedTime}ms`);
        }
    });
}
exports.refreshSeenSequences = refreshSeenSequences;
async function registerEventListeners(app, redisPool, opts, prefix) {
    async function markVaaSeen(vaa) {
        let redis;
        try {
            redis = await redisPool.acquire();
        }
        catch (error) {
            opts.logger?.error(`Failed to acquire redis client while trying to mark vaa seen.`, error);
            return;
        }
        try {
            await (0, storage_1.markVaaAsSeen)(redis, vaa.id, prefix);
        }
        catch (error) {
            opts.logger?.error(`Failed to mark VAA ass seen ${vaa.id.sequence}. Vaa will probably be reprocessed. Error: `, error);
        }
        finally {
            await redisPool.release(redis);
        }
    }
    app.addListener(application_1.RelayerEvents.Added, markVaaSeen);
    app.addListener(application_1.RelayerEvents.Skipped, markVaaSeen);
}
exports.registerEventListeners = registerEventListeners;
function scanForSequenceLeaps(seenSequences) {
    const missing = [];
    let idx = 0;
    let nextSeen = seenSequences[0];
    for (let seq = seenSequences[0]; seq < seenSequences[seenSequences.length - 1]; seq++) {
        if (seq === nextSeen) {
            nextSeen = seenSequences[++idx];
            continue;
        }
        missing.push(seq);
    }
    return missing;
}
//# sourceMappingURL=check.js.map