"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.runMissedVaaCheck = exports.spawnMissedVaaWorker = void 0;
const ioredis_1 = require("ioredis");
const generic_pool_1 = require("generic-pool");
const application_1 = require("../../application");
const utils_1 = require("../../utils");
const metrics_1 = require("./metrics");
const helpers_1 = require("./helpers");
const check_1 = require("./check");
const storage_1 = require("./storage");
const helpers_2 = require("./helpers");
const DEFAULT_PREFIX = "MissedVaaWorkerV3";
let metrics;
async function spawnMissedVaaWorker(app, opts) {
    opts.wormholeRpcs = opts.wormholeRpcs ?? application_1.defaultWormholeRpcs[app.env];
    if (!metrics) {
        metrics = opts.registry ? (0, metrics_1.initMetrics)(opts.registry) : {};
    }
    const redisPool = createRedisPool(opts);
    if (!app.filters.length) {
        opts.logger?.warn("Missed VAA Worker: No filters found, retrying in 100ms...");
        setTimeout(() => {
            spawnMissedVaaWorker(app, opts);
        }, 100);
        return;
    }
    const filters = app.filters.map(filter => {
        return {
            emitterChain: filter.emitterFilter.chainId,
            emitterAddress: filter.emitterFilter.emitterAddress,
        };
    });
    const prefix = opts.storagePrefix || DEFAULT_PREFIX;
    if (opts.storagePrefix) {
        await (0, check_1.refreshSeenSequences)(redisPool, filters, opts);
    }
    (0, check_1.registerEventListeners)(app, redisPool, opts, prefix);
    while (true) {
        opts.logger?.info(`Missed VAA middleware run starting...`);
        await (0, utils_1.mapConcurrent)(filters, filter => redisPool.use(async (redis) => {
            const startTime = Date.now();
            const filterLogger = opts.logger?.child({
                emitterChain: filter.emitterChain,
                emitterAddress: filter.emitterAddress,
            });
            try {
                const results = await runMissedVaaCheck(filter, redis, app.processVaa.bind(app), opts, prefix, filterLogger);
                (0, helpers_1.updateMetrics)(metrics, filter, startTime, false, results.missedVaas, results.sequenceStats);
            }
            catch (error) {
                filterLogger?.error(`Error checking for missed vaas for filter: ${JSON.stringify(filter)}}`, error);
                (0, helpers_1.updateMetrics)(metrics, filter, startTime, true);
            }
        }));
        await (0, utils_1.sleep)(opts.checkInterval || 30_000);
    }
}
exports.spawnMissedVaaWorker = spawnMissedVaaWorker;
async function runMissedVaaCheck(filter, redis, processVaa, opts, storagePrefix, logger) {
    const { emitterChain, emitterAddress } = filter;
    const previousSafeSequence = await (0, storage_1.tryGetLastSafeSequence)(redis, storagePrefix, emitterChain, emitterAddress, logger);
    const missedVaas = await (0, check_1.checkForMissedVaas)(filter, redis, processVaa, opts, storagePrefix, previousSafeSequence, logger);
    // TODO: this is an ugly way to handle the error
    const failedToFetchSequencesOrError = await (0, storage_1.tryGetExistingFailedSequences)(redis, filter, storagePrefix);
    if (!Array.isArray(failedToFetchSequencesOrError)) {
        logger?.error(`Failed to get existing failed sequences from redis. Error: `, failedToFetchSequencesOrError);
    }
    else if (failedToFetchSequencesOrError.length) {
        logger?.warn(`Found sequences that we failed to get from wormhole-rpc. Sequences: ` +
            JSON.stringify(failedToFetchSequencesOrError));
    }
    else {
        logger?.debug("No previous failed sequences found.");
    }
    const failedToFetchSequences = Array.isArray(failedToFetchSequencesOrError)
        ? failedToFetchSequencesOrError
        : null;
    const sequenceStats = (0, helpers_2.calculateSequenceStats)(missedVaas, failedToFetchSequences, previousSafeSequence?.toString());
    const { lastSafeSequence, lastSeenSequence, firstSeenSequence } = sequenceStats;
    if (!previousSafeSequence ||
        previousSafeSequence?.toString() !== String(lastSafeSequence)) {
        logger?.debug(`No missing sequences found up to sequence ${lastSafeSequence}. Setting as last sequence`);
        await (0, storage_1.trySetLastSafeSequence)(redis, storagePrefix, emitterChain, emitterAddress, lastSafeSequence);
    }
    const vaasFound = missedVaas.missingSequences.length + missedVaas.lookAheadSequences.length;
    logger?.info(`Finished missed VAAs check. Found: ${vaasFound}. Enable debug to see a log with precise results.`);
    logger?.debug(`Finished missed vaas check. Results: ${JSON.stringify({
        missedVaas,
        lastSafeSequence,
        lastSeenSequence,
        firstSeenSequence,
    })}}`);
    return { missedVaas, sequenceStats };
}
exports.runMissedVaaCheck = runMissedVaaCheck;
function createRedisPool(opts) {
    const factory = {
        create: async function () {
            const redis = opts.redisCluster
                ? new ioredis_1.default.Cluster(opts.redisClusterEndpoints, opts.redisCluster)
                : new ioredis_1.default(opts.redis);
            // TODO: metrics.missed_vaa_redis_open_connections.inc();
            return redis;
        },
        destroy: async function (redis) {
            // TODO: metrics.missed_vaa_redis_open_connections.dec();
        },
    };
    const poolOpts = {
        min: 5,
        max: 15,
        autostart: true,
    };
    return (0, generic_pool_1.createPool)(factory, poolOpts);
}
//# sourceMappingURL=worker.js.map