"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createWalletToolbox = void 0;
const wh = require("@certusone/wormhole-sdk");
const bs58 = require("bs58");
const ethers_1 = require("ethers");
const solana = require("@solana/web3.js");
const sui_js_1 = require("@mysten/sui.js");
const proto_signing_1 = require("@cosmjs/proto-signing");
async function createWalletToolbox(providers, privateKey, chainId) {
    if (wh.isEVMChain(chainId)) {
        return createEVMWalletToolBox(providers, privateKey, chainId);
    }
    switch (chainId) {
        case wh.CHAIN_ID_SOLANA:
            let secretKey;
            try {
                secretKey = bs58.decode(privateKey);
            }
            catch (e) {
                secretKey = new Uint8Array(JSON.parse(privateKey));
            }
            return createSolanaWalletToolBox(providers, secretKey);
        case wh.CHAIN_ID_SUI:
            const secret = Buffer.from(privateKey, "base64");
            return createSuiWalletToolBox(providers, secret);
        case wh.CHAIN_ID_SEI:
            const seiPkBuf = Buffer.from(privateKey, "hex");
            return createSeiWalletToolBox(providers, seiPkBuf);
    }
}
exports.createWalletToolbox = createWalletToolbox;
function createEVMWalletToolBox(providers, privateKey, chainId) {
    const wallet = new ethers_1.ethers.Wallet(privateKey, providers.evm[chainId][0]);
    return {
        ...providers,
        wallet: wallet,
        async getBalance() {
            const b = await wallet.getBalance();
            return b.toString();
        },
        address: wallet.address,
    };
}
function createSolanaWalletToolBox(providers, privateKey) {
    return {
        ...providers,
        wallet: {
            conn: providers.solana[0],
            payer: solana.Keypair.fromSecretKey(privateKey),
        },
        async getBalance() {
            return "TODO";
        },
        address: "TODO",
    };
}
function createSuiWalletToolBox(providers, secret) {
    const keyPair = sui_js_1.Ed25519Keypair.fromSecretKey(secret);
    const suiProvider = providers.sui[0];
    const wallet = new sui_js_1.RawSigner(keyPair, suiProvider);
    const address = keyPair.getPublicKey().toSuiAddress();
    return {
        ...providers,
        wallet,
        async getBalance() {
            const b = await suiProvider.getBalance({
                owner: address,
            });
            return b.totalBalance.toString();
        },
        address: address,
    };
}
async function createSeiWalletToolBox(providers, privateKey) {
    const seiWallet = await proto_signing_1.DirectSecp256k1Wallet.fromKey(privateKey, "sei");
    const [seiAccount] = await seiWallet.getAccounts();
    const seiProvider = providers.sei[0];
    return {
        ...providers,
        wallet: seiWallet,
        address: seiAccount.address,
        async getBalance() {
            const b = await seiProvider.getBalance(seiAccount.address, "usei");
            return b.amount;
        },
    };
}
//# sourceMappingURL=walletToolBox.js.map